/*
 * KLE SERVER
 * Copyright (C) 2004-2008 Prosum <contact@prosum.fr>
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/select.h>


#ifndef false
#define false 0
#define true -1
#endif 

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

#include "kle.h"

void sendMRSignal(kleScreenPtr screen, RegionPtr modRegion)
{       
    kleClientPtr cl = screen->kleClientHead;
    while (cl) {
        LOCK(cl->updateMutex);
        REGION_UNION(&cl->modifiedRegion, &cl->modifiedRegion, modRegion );
        UNLOCK(cl->updateMutex);
        SIGNAL(cl->updateCond);
        cl = nextClient(cl);
    }
}

void printRegion (RegionPtr reg, char *name)
{
    DBGP ("Region %s:\n", name);
    DBGP ("  extents: %d %d %d %d\n", reg->extents.x1,reg->extents.y1, 
                                            reg->extents.x2, reg->extents.y2);
    if (reg->data)
        DBGP ("  data: size=%d, numRects=%d\n", (int)reg->data->size, (int)reg->data->numRects);
}


/* Thread */   
static void shsendto (int signal , siginfo_t  si , void* context)
{
    kleLog("SIGSEGV in sendToClient at address 0x%x\n", si.si_addr);
    exit(1);
}

static void *sendToClient(void *data)
{       
    kleClientPtr    cl = (kleClientPtr) data;
    RegionRec       updateRegion;
    RegionPtr       reqReg = &cl->requestedRegion;
    RegionPtr       modReg = &cl->modifiedRegion; 
    RegionPtr       updReg = &updateRegion;
    struct sigaction   act, oact;
    struct sigaction   m; 
    struct timeval t0, t1;
    Bool allowed = TRUE;

    /* capture segmentation fault signal */
    m.sa_flags = SA_SIGINFO;
    sigemptyset(&m.sa_mask); 
    m.sa_sigaction = (void *)shsendto;
    sigaction (SIGSEGV, &m, NULL);

    /* ignore broken pipe signal */
    act.sa_handler = SIG_IGN;           /* ignore */
    sigemptyset(&act.sa_mask); 
    act.sa_flags = 0;
    sigaction(SIGPIPE, &act, &oact);    /* catch broken pipe signal */

    /* Wait a while for Client Input start */
    usleep (5000);

    gettimeofday (&t0, NULL);

    LOCK(cl->updateMutex);
    while (1) {
        while ( (REGION_NIL(modReg) || REGION_NULL_EXTENTS(reqReg))
                && !cl->extensionData 
                && cl->sock >= 0) {
            WAIT (cl->updateCond, cl->updateMutex);        /* here updateMutex is released */
        }
        if (cl->sock == - 1) break;

        if (cl->extensionData)
            kleSendExtensionMsg (cl);

#if KDEVICE==IKMOD || KDEVICE==IKV2 || KDEVICE==IKDUAL
        if (cl->allowed == FALSE) {
            if (!REGION_NULL_EXTENTS(reqReg)) {
                kleSendPanelFrameBuffer(cl, sorryReg);
                REGION_EMPTY(reqReg);
            }
            allowed = FALSE;
        }
        else
#endif  
        {             
            if (!(REGION_NIL(modReg) || REGION_NULL_EXTENTS(reqReg))) {
                REGION_INIT(updReg, NullBox, 20);
                if (REGION_EXTENTS_COVERS(reqReg, modReg)) {
                    /* normal case */
                    REGION_COPY(updReg, modReg);    /* reqReg includes modReg */
                    REGION_EMPTY(modReg);
                }
                else {
                    REGION_COPY(updReg, reqReg);    /* modReg includes reqReg */
                    if (allowed) {
                        REGION_SUBTRACT(modReg, modReg, updReg);
                    }
                    else {

                        /* 
                         * just allowed : sleep 2 seconds not to see the current 
                         *  non authorized image because the KVM certainly did not 
                         *  get time to switch on the authorized port yet 
                         *  
                         */
                        UNLOCK(cl->updateMutex);
                        sleep (2);
                        LOCK(cl->updateMutex);
                        REGION_EMPTY(modReg);
                    }
                }
                allowed = TRUE;               

                /* Send update. */
                if (cl->sock ==  - 1) break;
                gettimeofday (&t0, NULL);
                kleSendFramebufferUpdate(cl, updReg);
                gettimeofday (&t1, NULL);
                timevalsub (&cl->timePerImage, &t1, &t0);
                REGION_EMPTY(reqReg);  
            }
        }
    }
    /*exit thread */
    UNLOCK(cl->updateMutex);
    sigaction(SIGPIPE, &oact, NULL);
    if (uniParam.securityLevel != NO_SECURITY)
        ERR_remove_state(0);
    pthread_exit (NULL);
    return NULL;
}

int clientInitSSL (kleClientPtr cl)
{
    int    err, error;
    int    one = 1;
    fd_set rfds;
    struct timeval tv = {30,0};
    
    FD_ZERO(&rfds);

    /*  Set socket in non blocking mode */
    if (fcntl(cl->sock, F_SETFL, O_NONBLOCK) < 0) {    
        kleLogPerror("fcntl error");
        goto error;
    }
    /* Select socket for input */
    FD_SET(cl->sock,&rfds);

    /* Timeout if the client does not speak SSL */
    if (select(cl->sock+1, &rfds, NULL, NULL, &tv) == 0) {
        kleLog("Peer rejected: Timeout\n");
        goto error;
    }
    /* Come back to blocking mode */
    if (fcntl(cl->sock, F_SETFL, 0) < 0) {    
        kleLogPerror("fcntl error");
        goto error;
    }

    err = SSL_accept(cl->ssl);
    if (err == 0){
        kleLog("Peer gave up\n");
        goto error;
    }                       
    else if (err == -1) {
        error = SSL_get_error (cl->ssl, err); 
        if (error != SSL_ERROR_WANT_READ) {
            kleLog ("Peer rejected: SSL handshake %d\n", error);
            goto error;
        }
    }
    else if (err < -1) {        
        kleLog("Peer rejected: SSL connection %d\n", SSL_get_error(cl->ssl, err));
//        intSslCheck (cl->ssl, err);
        goto error;    
    }

    if (uniParam.securityLevel == HIGH_SECURITY) { 
        err = postConnectionCheck(cl->ssl, &cl->host);
        if (!cl->host) {
            kleLog ("Peer rejected: couldn't be identified\n");
            goto error;
        }
        if (err != X509_V_OK) {
            kleLog ("Peer rejected: wrong certificate: %s\n", X509_verify_cert_error_string(err));
            goto error;
        }
    }
    /* set socket options */
    if (setsockopt(cl->sock, IPPROTO_TCP, TCP_NODELAY, (char *)&one, sizeof(one)) < 0) {
        kleLogPerror("setsockopt error");
        goto error;
    }       
    if (writeToClient(cl, protocolVersion, SIZE_PROTO_VERSION-1) < 0) {
        kleLog ("First writeToClient didn't work\n");
        kleCloseClient (cl);
        goto error;
    } 

    requestMouseResynch (cl->screen);
    return 0;

error:
    /* Unlink the client from the client list */
    /* free everything */
    SSL_free(cl->ssl);
    kleClientUnlink (cl);
    return -1;

}

/* Thread */
static void shrecfrom (int signal , siginfo_t  si , void* context)
{
    kleLog("SIGSEGV in receiveFromClient at address 0x%x\n", si.si_addr);
    exit(1);
}
static void *receiveFromClient(void *data)
{       
    kleClientPtr cl = (kleClientPtr)data;
    pthread_t output_thread;
    struct sched_param sp;
    pthread_attr_t attrj;
    struct sigaction act, oact;
    struct sigaction   m; 

    /* capture segmentation fault signal */
    m.sa_flags = SA_SIGINFO;
    sigemptyset(&m.sa_mask); 
    m.sa_sigaction = (void *)shrecfrom;
    sigaction (SIGSEGV, &m, NULL);

     /* ignore broken pipe signal */     
    act.sa_handler = SIG_IGN; 
    sigemptyset(&act.sa_mask); 
    act.sa_flags = 0;
    sigaction(SIGPIPE, &act, &oact);

    if (uniParam.securityLevel != NO_SECURITY) {
        if (clientInitSSL(cl)){
            ERR_remove_state(0);
            pthread_exit(NULL);
            return NULL;
        }
    }    

    pthread_attr_init (&attrj);
    pthread_attr_setdetachstate (&attrj, PTHREAD_CREATE_JOINABLE);
    pthread_attr_setschedpolicy (&attrj, SCHED_FIFO);

    /* set priority */
    pthread_attr_getschedparam (&attrj, &sp);
    sp.sched_priority = sched_get_priority_min (SCHED_FIFO) + 10;
    pthread_attr_setschedparam (&attrj, &sp);

    pthread_create(&output_thread, &attrj, sendToClient, (void *) cl);
    
    pthread_attr_destroy (&attrj);

    while (1) {
        kleProcessClientMessage(cl);
        if (cl->sock == - 1) {
            /* Client has disconnected. */
            break;
        }
    }
    /* Await output thread stop */
    SIGNAL(cl->updateCond);
    pthread_join (output_thread, NULL);
    if (uniParam.securityLevel != NO_SECURITY) {
        SSL *ssl = cl->ssl;
        /* Free SSL object */
//        if ((SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN))
//            SSL_shutdown(ssl);
//        else
//            SSL_clear(ssl);
//
        SSL_shutdown (ssl);
        SSL_free(ssl);
        ERR_remove_state(0);
    }
    kleClientConnectionClosed(cl);
    sigaction(SIGPIPE, &oact, NULL);
    pthread_exit(NULL);
    return NULL;
}



/* Thread */
void *waitForClients(void *ic)
{
    kleScreenPtr screen = (kleScreenPtr)ic;
    int             clientFd;
    pthread_t       clientThread;
    kleClientPtr    cl;
    int             one = 1;
    char            *host;
    struct sockaddr_in peer;
    socklen_t len = sizeof (peer);
    struct          sched_param sp;
    pthread_attr_t  attrj;

    pthread_attr_init (&attrj);
    pthread_attr_setschedpolicy (&attrj, SCHED_FIFO);
    pthread_attr_getschedparam (&attrj, &sp);
    sp.sched_priority = sched_get_priority_min (SCHED_FIFO) + 9;
    pthread_attr_setschedparam (&attrj, &sp);

    if(screen->klePort > 0) {
        kleLog("Listening for VNC connections on TCP port %d\n", screen->klePort);
        if ((screen->kleListenSock = listenPort(screen->klePort)) < 0) {
            kleLogPerror("Socket error, closing");
            exit(1);
        }
    }
    while ((clientFd = accept(screen->kleListenSock, 
                                        (struct sockaddr *)&peer, &len)) >= 0) {
        host = strdup(inet_ntoa(peer.sin_addr));
        if (!host) {
            kleLog ("Error: couldn't store client IP address\n"); 
            break;
        }
        /*  set socket in non blocking mode */
//            if (fcntl(clientFd, F_SETFL, O_NONBLOCK) < 0) {    
//                kleLogPerror("fcntl error");
//                break;
//            }
        
        /* set socket options */
        if (setsockopt(clientFd, IPPROTO_TCP, TCP_NODELAY, 
                                            (char *)&one, sizeof(one)) < 0) {
            kleLogPerror("setsockopt error");
            break;
        }
        cl = kleNewClient(screen, NULL, NULL, clientFd, host);
        if (!cl) {
            kleLog ("Couldn't create client structure\n");
            free (host);
            break;
        }
        if (writeToClient(cl, protocolVersion, SIZE_PROTO_VERSION-1) < 0) {
            kleLog ("First writeToClient didn't work\n");
            kleCloseClient (cl);
            free (host);
            free (cl);
            continue;
        } 
        requestMouseResynch (screen); 
        pthread_create(&clientThread, &attrj, receiveFromClient, (void *)cl);
        len = sizeof (peer);
    }
    pthread_attr_destroy (&attrj);
    kleLog ("Stopping the listener\n");
    exit (-1);         /* stop the server */
    return NULL;       /* for compiler */
}

/* Thread */
void *waitForClientsSSL(void *ic)
{
    kleScreenPtr screen = (kleScreenPtr)ic;
    int             clientFd;
    pthread_t       clientThread;
    kleClientPtr    cl;
    char            *host;
    BIO             *bio;
    SSL             *ssl;
    char            port_str[20];
    struct          sched_param sp;
    pthread_attr_t  attrj;
    struct sockaddr_in addr;
    socklen_t       add_len = sizeof(addr);

    pthread_attr_init (&attrj);
    pthread_attr_setschedpolicy (&attrj, SCHED_FIFO);
    pthread_attr_getschedparam (&attrj, &sp);
    sp.sched_priority = sched_get_priority_min (SCHED_FIFO) + 9;
    pthread_attr_setschedparam (&attrj, &sp);

    sprintf (port_str, "%d", screen->klePort); 
    screen->acc = BIO_new_accept(port_str);
    if (!screen->acc) {
        kleLog("Error creating server socket\n");
        goto error;
    }

    if (BIO_do_accept(screen->acc) <= 0) {
        kleLog("Error binding server socket\n");
        goto error;
    }

    for (;;) {
        host = NULL;
        cl = NULL;
        ssl = NULL;
        if (BIO_do_accept(screen->acc) <= 0) {
            kleLog("Error accepting connection\n");
            break;
        }
        bio = BIO_pop(screen->acc);
        if (!(ssl = SSL_new(ctx))) {
            kleLog("Error creating SSL context\n");
            break;
        }

        clientFd = BIO_get_fd (bio, NULL);
        
        if ( getpeername(clientFd, (struct sockaddr*)&addr, &add_len) != 0 )
            kleLogPerror("getpeername() failed");
        
        kleLog("Peer %s tries to connect\n", inet_ntoa(addr.sin_addr));
        
        SSL_set_accept_state(ssl);    /* server mode */
        SSL_set_bio(ssl, bio, bio);

        cl = kleNewClient(screen, bio, ssl, clientFd, host);
        if (!cl) {
            kleLog ("Couldn't create client structure\n");
            break;
        }
        pthread_create(&clientThread, &attrj, receiveFromClient, (void *)cl);
    
    }

    pthread_attr_destroy (&attrj);

    if (host)
        free(host);
    if (cl)
        free(cl);
    if (ssl) 
        SSL_free(ssl);

    ERR_remove_state(0);

error:
    if (screen->acc) {
        BIO_free(screen->acc);
        screen->acc = NULL;
    }

    kleLog ("Stopping the SSL listener\n");
    exit (-1);         /* stop the server */
    return NULL;       /* for compiler */
}

void requestScreenClear(kleClientPtr cl)
{
    kleScreenPtr screen =  cl->screen;
    screen->clearRequest = 1;
}

void initScreenComputers (kleScreenPtr screen)
{
    int i;

    memset (screen->cpus, sizeof (screen->cpus), 0);

    for (i =0 ; i< CPU_MAX; i++) {
        screen->cpus[i].digi_gain = 0x4e;
        screen->cpus[i].digi_offset = 0x2d;
        screen->cpus[i].monit = &monitors[0];
    }
}

void kleScreenCleanup(kleScreenPtr screen)
{       
    kleClientPtr cl, nextcl;
    
    if(screen) {
        if (screen->kmqp)
            free(screen->kmqp);
        
        cl = screen->kleClientHead;
        while (cl != NULL) {
            nextcl = nextClient(cl);
            cl->auth.sessionTerminateCause = ADMIN_REBOOT; 
            kleCloseClient(cl);
            kleClientFreeMem (cl);
            cl = nextcl;
        }

        if (screen->colourMap.data.bytes)
            free(screen->colourMap.data.bytes);

        free(screen);
    }
}

#undef free

void *kleMalloc(int lbs)
{
    void *ptr= malloc(lbs);
    if (ptr == 0)
        kleLogPerror("kle: out of memory");
    return ptr;
}

void *kleRealloc(void *old, int lbs)
{
    void *ptr= realloc(old, lbs);
    if (ptr == 0)
        kleLogPerror("kle: out of memory");
    return ptr;
}

void *kleCalloc(int nmemb, int size)
{
    void *ptr= calloc(nmemb, size);
    if (ptr == 0)
        kleLogPerror("kle: out of memory");
    return ptr;
}


